@@ -1,16 +1,19 @@
 #!/usr/bin/env python
-# Copyright 2011 Google Inc. All Rights Reserved.
+# Copyright 2012 Google Inc. All Rights Reserved.
 
 """Bigquery Client library for Python."""
 
 
 
+import abc
 import datetime
+import hashlib
 import itertools
 import json
 import logging
 import os
 import pkgutil
+import random
 import re
 import string
 import sys
@@ -24,6 +27,7 @@ from apiclient import http as http_request
 from apiclient import model
 import httplib2
 
+
 # To configure apiclient logging.
 import gflags as flags
 
@@ -57,9 +61,10 @@ def _ApplyParameters(config, **kwds):
 
   Note this does not remove entries that are set to None, however.
 
+  kwds: A dict of keys and values to set in the config.
+
   Args:
     config: A configuration dict.
-    kwds: A dict of keys and values to set in the config.
   """
   config.update((_ToLowerCamel(k), v) for k, v in kwds.iteritems()
                 if v is not None)
@@ -93,6 +98,8 @@ def ConfigurePythonLogger(apilog=None):
     flags.FLAGS.dump_request_response = True
 
 
+
+
 class BigqueryError(Exception):
 
   @staticmethod
@@ -143,6 +150,9 @@ class BigqueryError(Exception):
     if reason == 'termsOfServiceNotAccepted':
       return BigqueryTermsOfServiceError(
           message, error, error_ls, job_ref=job_ref)
+    if reason == 'backendError':
+      return BigqueryBackendError(
+          message, error, error_ls, job_ref=job_ref)
     # We map the less interesting errors to BigqueryServiceError.
     return BigqueryServiceError(message, error, error_ls, job_ref=job_ref)
 
@@ -211,6 +221,11 @@ class BigqueryTermsOfServiceError(BigqueryAccessDeniedError):
   pass
 
 
+class BigqueryBackendError(BigqueryServiceError):
+  """A backend error typically corresponding to retriable HTTP 503 failures."""
+  pass
+
+
 class BigqueryClientError(BigqueryError):
   """Invalid use of BigqueryClient."""
   pass
@@ -263,15 +278,84 @@ class BigqueryHttp(http_request.HttpRequest):
     try:
       return super(BigqueryHttp, self).execute(**kwds)
     except apiclient.errors.HttpError, e:
+      # TODO(user): Remove this when apiclient supports logging
+      # of error responses.
+      self._model._log_response(e.resp, e.content)  # pylint:disable-msg=W0212
       if e.resp.get('content-type', '').startswith('application/json'):
-        # TODO(user): Remove this when apiclient supports logging
-        # of error responses.
-        self._model._log_response(e.resp, e.content)  # pylint:disable-msg=W0212
         BigqueryClient.RaiseError(json.loads(e.content))
       else:
         raise BigqueryCommunicationError(
-            'Could not connect with BigQuery server, http response status: ' +
-            e.resp.get('status', '(unexpected)'))
+            ('Could not connect with BigQuery server.\n'
+             'Http response status: %s\n'
+             'Http response content:\n%s') % (
+                 e.resp.get('status', '(unexpected)'), e.content))
+
+
+class JobIdGenerator(object):
+  """Base class for job id generators."""
+  __metaclass__ = abc.ABCMeta
+
+  @abc.abstractmethod
+  def Generate(self, job_configuration):
+    """Generates a job_id to use for job_configuration."""
+
+
+class JobIdGeneratorNone(JobIdGenerator):
+  """Job id generator that returns None, letting the server pick the job id."""
+
+  def Generate(self, unused_config):
+    return None
+
+
+class JobIdGeneratorRandom(JobIdGenerator):
+  """Generates random job ids."""
+
+  def Generate(self, unused_config):
+    return 'bqjob_r%08x_%016x' % (random.randint(0, sys.maxint),
+                                  int(time.time() * 1000))
+
+
+class JobIdGeneratorFingerprint(JobIdGenerator):
+  """Generates job ids that uniquely match the job config."""
+
+  def _Hash(self, config, sha1):
+    """Computes the sha1 hash of a dict."""
+    keys = config.keys()
+    # Python dict enumeration ordering is random. Sort the keys
+    # so that we will visit them in a stable order.
+    keys.sort()
+    for key in keys:
+      sha1.update('%s' % (key,))
+      v = config[key]
+      if isinstance(v, dict):
+        logging.info('Hashing: %s...', key)
+        self._Hash(v, sha1)
+      elif isinstance(v, list):
+        logging.info('Hashing: %s ...', key)
+        for inner_v in v:
+          self._Hash(inner_v, sha1)
+      else:
+        logging.info('Hashing: %s:%s', key, v)
+        sha1.update('%s' % (v,))
+
+  def Generate(self, config):
+    s1 = hashlib.sha1()
+    self._Hash(config, s1)
+    job_id = 'bqjob_c%s' % (s1.hexdigest(),)
+    logging.info('Fingerprinting: %s:\n%s', config, job_id)
+    return job_id
+
+
+class JobIdGeneratorIncrementing(JobIdGenerator):
+  """Generates job ids that increment each time we're asked."""
+
+  def __init__(self, inner):
+    self._inner = inner
+    self._retry = 0
+
+  def Generate(self, config):
+    self._retry += 1
+    return '%s_%d' % (self._inner.Generate(config), self._retry)
 
 
 class BigqueryClient(object):
@@ -321,6 +405,7 @@ class BigqueryClient(object):
         'trace': None,
         'sync': True,
         'wait_printer_factory': BigqueryClient.TransitionWaitPrinter,
+        'job_id_generator': JobIdGeneratorIncrementing(JobIdGeneratorRandom()),
         }
     for flagname, default in default_flag_values.iteritems():
       if not hasattr(self, flagname):
@@ -328,11 +413,21 @@ class BigqueryClient(object):
     if self.dataset_id and not self.project_id:
       raise ValueError('Cannot set dataset_id without project_id')
 
+  def GetHttp(self):
+    """Returns the httplib2 Http to use."""
+    http = httplib2.Http()
+    return http
+
+  def GetDiscoveryUrl(self):
+    """Returns the url to the discovery document for bigquery."""
+    discovery_url = self.api + '/discovery/v1/apis/{api}/{apiVersion}/rest'
+    return discovery_url
+
   @property
   def apiclient(self):
     """Return the apiclient attached to self."""
     if self._apiclient is None:
-      http = self.credentials.authorize(httplib2.Http())
+      http = self.credentials.authorize(self.GetHttp())
       bigquery_model = BigqueryModel(self.trace)
       bigquery_http = BigqueryHttp.Factory(
           bigquery_model)
@@ -346,20 +441,19 @@ class BigqueryClient(object):
         except IOError:
           discovery_document = None
       if discovery_document is None:
-        discovery_url = '%s/discovery/v1/apis/{api}/{apiVersion}/rest' % (
-            self.api,)
         try:
           self._apiclient = discovery.build(
               'bigquery', self.api_version, http=http,
-              discoveryServiceUrl=discovery_url,
+              discoveryServiceUrl=self.GetDiscoveryUrl(),
               model=bigquery_model,
               requestBuilder=bigquery_http)
         except (httplib2.ServerNotFoundError, apiclient.errors.HttpError), e:
           # We can't find the specified server.
           raise BigqueryCommunicationError(
-              'Cannot contact server. Please try again. Error: %s' % (str(e),))
+              'Cannot contact server. Please try again.\nError: %s'
+              '\nContent: %s' % (str(e), e.content))
         except apiclient.errors.UnknownApiNameOrVersion, e:
-          # We can't resolve the discovery_url for the given server.
+          # We can't resolve the discovery url for the given server.
           raise BigqueryCommunicationError(
               'Invalid API name or version: %s' % (str(e),))
       else:
@@ -610,21 +704,29 @@ class BigqueryClient(object):
 
   def ReadTableRows(self, table_dict, max_rows=sys.maxint):
     """Read at most max_rows rows from a table."""
+    page_token = None
     rows = []
     while len(rows) < max_rows:
       data = self.apiclient.tabledata().list(
-          maxResults=min(10000, max_rows - len(rows)),
-          startIndex=len(rows), **table_dict).execute()
+          maxResults=max_rows - len(rows),
+          pageToken=page_token,
+          **table_dict).execute()
+      page_token = data.get('pageToken', None)
       max_rows = min(max_rows, int(data['totalRows']))
       more_rows = data.get('rows', [])
       for row in more_rows:
         rows.append([entry.get('v', '') for entry in row.get('f', [])])
+      if not page_token and len(rows) != max_rows:
+        raise BigqueryInterfaceError(
+            'PageToken missing for %r' % (
+                ApiClientHelper.TableReference.Create(**table_dict),))
       if not more_rows and len(rows) != max_rows:
         raise BigqueryInterfaceError(
             'Not enough rows returned by server for %r' % (
                 ApiClientHelper.TableReference.Create(**table_dict),))
     return rows
 
+
   def ReadSchemaAndRows(self, table_dict, max_rows=sys.maxint):
     """Convenience method to get the schema and rows from a table.
 
@@ -974,7 +1076,8 @@ class BigqueryClient(object):
         BigqueryClient.ConstructObjectReference, self.ListJobs(**kwds))
 
   def ListJobs(self, reference=None,
-               max_results=None, state_filter=None):
+               max_results=None, state_filter=None,
+               all_users=None):
     """Return a list of jobs.
 
     Args:
@@ -982,6 +1085,8 @@ class BigqueryClient(object):
       max_results: The maximum number of jobs to return.
       state_filter: A single state filter or a list of filters to
         apply. If not specified, no filtering is applied.
+     all_users: Whether to list jobs for all users of the project. Requesting
+       user must be an owner of the project to list all jobs.
 
     Returns:
       A list of jobs.
@@ -989,14 +1094,14 @@ class BigqueryClient(object):
     reference = self._NormalizeProjectReference(reference)
     _Typecheck(reference, ApiClientHelper.ProjectReference, method='ListJobs')
     request = self._PrepareListRequest(reference, max_results, None)
-    request['projection'] = 'full'
     if state_filter is not None:
       # The apiclient wants enum values as lowercase strings.
       if isinstance(state_filter, basestring):
         state_filter = state_filter.lower()
       else:
         state_filter = [s.lower() for s in state_filter]
-      request['stateFilter'] = state_filter
+    _ApplyParameters(request, projection='full',
+                     state_filter=state_filter, all_users=all_users)
     jobs = self.apiclient.jobs().list(**request).execute()
     return jobs.get('jobs', [])
 
@@ -1283,11 +1388,14 @@ class BigqueryClient(object):
         self.project_id is used.
       upload_file: A file to include as a media upload to this request.
         Only valid on job requests that expect a media upload file.
-      job_id: A unique job_id to use for this job. If None, the server
-        will create a unique job_id for this request.
+      job_id: A unique job_id to use for this job. If a
+        JobIdGenerator, a job id will be generated from the job configuration.
+        If None, a unique job_id will be created for this request.
 
     Returns:
-      The job resource returned from the insert job request.
+      The job resource returned from the insert job request. If there is an
+      error, the jobReference field will still be filled out with the job
+      reference used in the request.
 
     Raises:
       BigqueryClientConfigurationError: if project_id and
@@ -1302,13 +1410,22 @@ class BigqueryClient(object):
       configuration['properties'] = dict(
           prop.partition('=')[0::2] for prop in self.job_property)
     job_request = {'configuration': configuration}
-    if job_id:
-      job_request['jobReference'] = {'jobId': job_id}
+
+    # Use the default job id generator if no job id was supplied.
+    job_id = job_id or self.job_id_generator
+
+    if isinstance(job_id, JobIdGenerator):
+      job_id = job_id.Generate(configuration)
+
+    if job_id is not None:
+      job_reference = {'jobId': job_id, 'projectId': project_id}
+      job_request['jobReference'] = job_reference
     media_upload = ''
     if upload_file:
+      resumable = True
       media_upload = http_request.MediaFileUpload(
           filename=upload_file, mimetype='application/octet-stream',
-          resumable=True)
+          resumable=resumable)
     result = self.apiclient.jobs().insert(
         body=job_request, media_body=media_upload,
         projectId=project_id).execute()
@@ -1318,8 +1435,9 @@ class BigqueryClient(object):
                           upload_file=None, job_id=None):
     result = self.StartJob(configuration, project_id=project_id,
                            upload_file=upload_file, job_id=job_id)
-    job_reference = BigqueryClient.ConstructObjectReference(result)
-    result = self.WaitJob(job_reference)
+    if result['status']['state'] != 'DONE':
+      job_reference = BigqueryClient.ConstructObjectReference(result)
+      result = self.WaitJob(job_reference)
     return self.RaiseIfJobError(result)
 
   def ExecuteJob(self, configuration, sync=None,
@@ -1441,6 +1559,9 @@ class BigqueryClient(object):
       except BigqueryCommunicationError, e:
         # Communication errors while waiting on a job are okay.
         logging.warning('Transient error during job status check: %s', e)
+      except BigqueryBackendError, e:
+        # Temporary server errors while waiting on a job are okay.
+        logging.warning('Transient error during job status check: %s', e)
       for _ in xrange(waits.next()):
         current_wait = time.time() - start_time
         printer.Print(job_reference.jobId, current_wait, current_status)
@@ -1491,9 +1612,17 @@ class BigqueryClient(object):
     job = self.Query(**new_kwds)
     return self.ReadTableRows(job['configuration']['query']['destinationTable'])
 
-  def Query(self, query, destination_table=None,
-            create_disposition=None, write_disposition=None,
-            priority=None, **kwds):
+  def Query(self, query,
+            destination_table=None,
+            create_disposition=None,
+            write_disposition=None,
+            priority=None,
+            preserve_nulls=None,
+            allow_large_results=None,
+            dry_run=None,
+            use_cache=None,
+            **kwds):
+    # pylint:disable-msg=g-doc-args
     """Execute the given query, returning the created job.
 
     The job will execute synchronously if sync=True is provided as an
@@ -1509,6 +1638,15 @@ class BigqueryClient(object):
           the destination_table.
       priority: Optional. Priority to run the query with. Either
           'INTERACTIVE' (default) or 'BATCH'.
+      preserve_nulls: Optional. Indicates whether to preserve nulls in input
+          data. Temporary flag; will be removed in a future version.
+      allow_large_results: Enables larger destination table sizes.
+      dry_run: Optional. Indicates whether the query will only be validated and
+          return processing statistics instead of actually running.
+      use_cache: Optional. Whether to use the query cache. If create_disposition
+          is CREATE_NEVER, will only run the query if the result is already
+          cached. Caching is best-effort only and you should not make
+          assumptions about whether or how long a query result will be cached.
       **kwds: Passed on to self.ExecuteJob.
 
     Raises:
@@ -1529,18 +1667,23 @@ class BigqueryClient(object):
         raise BigqueryError('Invalid value %s for destination_table: %s' % (
             destination_table, e))
       query_config['destinationTable'] = dict(reference)
-    if priority is not None:
-      query_config['priority'] = priority
     _ApplyParameters(
-        query_config, create_disposition=create_disposition,
-        write_disposition=write_disposition)
-    return self.ExecuteJob({'query': query_config}, **kwds)
+        query_config,
+        allow_large_results=allow_large_results,
+        create_disposition=create_disposition,
+        preserve_nulls=preserve_nulls,
+        priority=priority,
+        write_disposition=write_disposition,
+        use_query_cache=use_cache)
+    request = {'query': query_config}
+    _ApplyParameters(request, dry_run=dry_run)
+    return self.ExecuteJob(request, **kwds)
 
   def Load(self, destination_table_reference, source,
            schema=None, create_disposition=None, write_disposition=None,
            field_delimiter=None, skip_leading_rows=None, encoding=None,
            quote=None, max_bad_records=None, allow_quoted_newlines=None,
-           source_format=None,
+           source_format=None, allow_jagged_rows=None,
            **kwds):
     """Load the given data into BigQuery.
 
@@ -1567,7 +1710,9 @@ class BigqueryClient(object):
       allow_quoted_newlines: Optional. Whether to allow quoted newlines in csv
           import data.
       source_format: Optional. Format of source data. May be "CSV",
-         "DATASTORE_BACKUP", or "NEWLINE_DELIMITED_JSON".
+          "DATASTORE_BACKUP", or "NEWLINE_DELIMITED_JSON".
+      allow_jagged_rows: Optional. Whether to allow missing trailing optional
+          columns in csv import data.
       **kwds: Passed on to self.ExecuteJob.
 
     Returns:
@@ -1589,10 +1734,12 @@ class BigqueryClient(object):
         skip_leading_rows=skip_leading_rows, encoding=encoding,
         quote=quote, max_bad_records=max_bad_records,
         source_format=source_format,
-        allow_quoted_newlines=allow_quoted_newlines)
+        allow_quoted_newlines=allow_quoted_newlines,
+        allow_jagged_rows=allow_jagged_rows)
     return self.ExecuteJob(configuration={'load': load_config},
                            upload_file=upload_file, **kwds)
 
+
   def Extract(self, source_table, destination_uri,
               print_header=None, field_delimiter=None,
               destination_format=None,
