@@ -1,5 +1,6 @@
 #!/usr/bin/env python
-# Copyright 2011 Google Inc. All Rights Reserved.
+#
+# Copyright 2012 Google Inc. All Rights Reserved.
 
 """Python script for interacting with BigQuery."""
 
@@ -8,6 +9,7 @@
 
 import cmd
 import codecs
+import datetime
 import httplib
 import json
 import os
@@ -26,6 +28,7 @@ import httplib2
 import oauth2client
 import oauth2client.client
 import oauth2client.file
+import oauth2client.gce
 import oauth2client.tools
 
 from google.apputils import app
@@ -44,8 +47,7 @@ flags.DEFINE_string(
 flags.DEFINE_string(
     'api',
     'https://www.googleapis.com',
-    'API endpoint to talk to.'
-    )
+    'API endpoint to talk to.')
 flags.DEFINE_string(
     'api_version', 'v2',
     'API version to use.')
@@ -90,6 +92,11 @@ flags.DEFINE_string(
     'A unique job_id to use for the request. If None, the server will create '
     'a unique job_id. Applies only to commands that launch jobs, such as cp, '
     'extract, link, load, and query. ')
+flags.DEFINE_boolean(
+    'fingerprint_job_id', False,
+    'Whether to use a job id that is derived from a fingerprint of the job '
+    'configuration. This will prevent the same job from running multiple times '
+    'accidentally.')
 flags.DEFINE_boolean(
     'quiet', False,
     'If True, ignore status updates while jobs are running.',
@@ -117,6 +124,11 @@ flags.DEFINE_multistring(
     'job_property', None,
     'Additional key-value pairs to include in the properties field of '
     'the job configuration')  # No period: Multistring adds flagspec suffix.
+flags.DEFINE_boolean(
+    'use_gce_service_account', False,
+    'Use this when running on a Google Compute Engine instance to use service '
+    'account credentials instead of stored credentials. For more information, '
+    'see: https://developers.google.com/compute/docs/authentication')
 flags.DEFINE_string(
     'service_account', '',
     'Use this service account email address for authorization. '
@@ -136,6 +148,7 @@ flags.DEFINE_string(
     'File to be used as a credential store for service accounts. '
     'Must be set if using a service account.')
 
+
 FLAGS = flags.FLAGS
 # These are long names.
 # pylint:disable-msg=C6409
@@ -144,10 +157,15 @@ ProjectReference = bigquery_client.ApiClientHelper.ProjectReference
 DatasetReference = bigquery_client.ApiClientHelper.DatasetReference
 TableReference = bigquery_client.ApiClientHelper.TableReference
 BigqueryClient = bigquery_client.BigqueryClient
+JobIdGeneratorIncrementing = bigquery_client.JobIdGeneratorIncrementing
+JobIdGeneratorRandom = bigquery_client.JobIdGeneratorRandom
+JobIdGeneratorFingerprint = bigquery_client.JobIdGeneratorFingerprint
 # pylint:enable-msg=C6409
 
 _CLIENT_USER_AGENT = 'bq/2.0'
-_CLIENT_SCOPE = 'https://www.googleapis.com/auth/bigquery'
+_CLIENT_SCOPE = [
+    'https://www.googleapis.com/auth/bigquery',
+]
 _CLIENT_ID = '977385342095.apps.googleusercontent.com'
 _CLIENT_INFO = {
     'client_id': _CLIENT_ID,
@@ -180,6 +198,13 @@ _DELIMITER_MAP = {
 
 
 
+def _ValidateGlobalFlags():
+  """Validate combinations of global flag values."""
+  if FLAGS.service_account and FLAGS.use_gce_service_account:
+    raise app.UsageError(
+        'Cannot specify both --service_account and --use_gce_service_account.')
+
+
 def _GetBigqueryRcFilename():
   """Return the name of the bigqueryrc file to use.
 
@@ -234,7 +259,14 @@ def _ResolveApiInfoFromFlags():
   return {'api': api, 'api_version': api_version}
 
 
+def _UseServiceAccount():
+  return bool(FLAGS.use_gce_service_account or FLAGS.service_account)
+
+
 def _GetServiceAccountCredentialsFromFlags(storage):  # pylint:disable-msg=W0613
+  if FLAGS.use_gce_service_account:
+    return oauth2client.gce.AppAssertionCredentials(_CLIENT_SCOPE)
+
   if not oauth2client.client.HAS_OPENSSL:
     raise app.UsageError(
         'BigQuery requires OpenSSL to be installed in order to use '
@@ -298,6 +330,11 @@ def _GetCredentialsFromOAuthFlow(storage):
 
 
 def _GetCredentialsFromFlags():
+  # In the case of a GCE service account, we can skip the entire
+  # process of loading from storage.
+  if FLAGS.use_gce_service_account:
+    return _GetServiceAccountCredentialsFromFlags(None)
+
 
   if FLAGS.service_account:
     credentials_getter = _GetServiceAccountCredentialsFromFlags
@@ -342,12 +379,74 @@ def _GetFormatterFromFlags(secondary_format='sparse'):
     return table_formatter.GetFormatter(secondary_format)
 
 
-def _PrintTable(client, table_dict, **extra_args):
-  fields, rows = client.ReadSchemaAndRows(table_dict, **extra_args)
-  formatter = _GetFormatterFromFlags(secondary_format='pretty')
-  formatter.AddFields(fields)
-  formatter.AddRows(rows)
-  formatter.Print()
+def _ExpandForPrinting(fields, rows, formatter):
+  """Expand entries that require special bq-specific formatting."""
+
+  def NormalizeTimestamp(entry):
+    try:
+      date = datetime.datetime.utcfromtimestamp(float(entry))
+      return date.strftime('%Y-%m-%d %H:%M:%S')
+    except ValueError:
+      return '<date out of range for display>'
+
+  column_normalizers = {}
+  for i, field in enumerate(fields):
+    if field['type'].upper() == 'TIMESTAMP':
+      column_normalizers[i] = NormalizeTimestamp
+
+  def NormalizeNone():
+    if isinstance(formatter, table_formatter.JsonFormatter):
+      return None
+    elif isinstance(formatter, table_formatter.CsvFormatter):
+      return ''
+    else:
+      return 'NULL'
+
+  def NormalizeEntry(i, entry):
+    if entry is None:
+      return NormalizeNone()
+    elif i in column_normalizers:
+      return column_normalizers[i](entry)
+    return entry
+
+  return ([NormalizeEntry(i, e) for i, e in enumerate(row)] for row in rows)
+
+
+def _PrintDryRunInfo(job):
+  num_bytes = job['statistics']['query']['totalBytesProcessed']
+  if FLAGS.format in ['prettyjson', 'json']:
+    _PrintFormattedJsonObject(job)
+  elif FLAGS.format == 'csv':
+    print num_bytes
+  else:
+    print (
+        'Query successfully validated. Assuming the tables are not modified, '
+        'running this query will process %s bytes of data.' % (num_bytes,))
+
+
+def _PrintFormattedJsonObject(obj):
+  if FLAGS.format == 'prettyjson':
+    print json.dumps(obj, sort_keys=True, indent=2)
+  else:
+    print json.dumps(obj, separators=(',', ':'))
+
+
+def _GetJobIdFromFlags():
+  """Returns the job id or job generator from the flags."""
+  if FLAGS.fingerprint_job_id and FLAGS.job_id:
+    raise app.UsageError(
+        'The fingerprint_job_id flag cannot be specified with the job_id '
+        'flag.')
+  if FLAGS.fingerprint_job_id:
+    return JobIdGeneratorFingerprint()
+  elif FLAGS.job_id is None:
+    return JobIdGeneratorIncrementing(JobIdGeneratorRandom())
+  elif FLAGS.job_id:
+    return FLAGS.job_id
+  else:
+    # User specified a job id, but it was empty. Let the
+    # server come up with a job id.
+    return None
 
 
 def _GetWaitPrinterFactoryFromFlags():
@@ -398,6 +497,61 @@ def _NormalizeFieldDelimiter(field_delimiter):
   return _DELIMITER_MAP.get(key, field_delimiter)
 
 
+class TablePrinter(object):
+  """Base class for printing a table, with a default implementation."""
+
+  def __init__(self, **kwds):
+    super(TablePrinter, self).__init__()
+    # Most extended classes will require state.
+    for key, value in kwds.iteritems():
+      setattr(self, key, value)
+
+  def PrintTable(self, fields, rows):
+    formatter = _GetFormatterFromFlags(secondary_format='pretty')
+    formatter.AddFields(fields)
+    rows = _ExpandForPrinting(fields, rows, formatter)
+    formatter.AddRows(rows)
+    formatter.Print()
+
+
+class Factory(object):
+  """Class encapsulating factory creation of BigqueryClient."""
+  _BIGQUERY_CLIENT_FACTORY = None
+
+  class ClientTablePrinter(object):
+    _TABLE_PRINTER = None
+
+    @classmethod
+    def GetTablePrinter(cls):
+      if cls._TABLE_PRINTER is None:
+        cls._TABLE_PRINTER = TablePrinter()
+      return cls._TABLE_PRINTER
+
+    @classmethod
+    def SetTablePrinter(cls, printer):
+      if not isinstance(printer, TablePrinter):
+        raise TypeError('Printer must be an instance of TablePrinter.')
+      cls._TABLE_PRINTER = printer
+
+  @classmethod
+  def GetBigqueryClientFactory(cls):
+    if cls._BIGQUERY_CLIENT_FACTORY is None:
+      cls._BIGQUERY_CLIENT_FACTORY = bigquery_client.BigqueryClient
+    return cls._BIGQUERY_CLIENT_FACTORY
+
+  @classmethod
+  def SetBigqueryClientFactory(cls, factory):
+    if not issubclass(factory, bigquery_client.BigqueryClient):
+      raise TypeError('Factory must be subclass of BigqueryClient.')
+    cls._BIGQUERY_CLIENT_FACTORY = factory
+
+
+def _PrintTable(client, table_dict, **extra_args):
+  fields, rows = client.ReadSchemaAndRows(table_dict, **extra_args)
+  printer = Factory.ClientTablePrinter.GetTablePrinter()
+  printer.PrintTable(fields, rows)
+
+
 class Client(object):
   """Class wrapping a singleton bigquery_client.BigqueryClient."""
   client = None
@@ -425,7 +579,8 @@ class Client(object):
     if FLAGS.discovery_file:
       with open(FLAGS.discovery_file) as f:
         client_args['discovery_document'] = f.read()
-    return BigqueryClient(credentials=credentials, **client_args)
+    bigquery_client_factory = Factory.GetBigqueryClientFactory()
+    return bigquery_client_factory(credentials=credentials, **client_args)
 
   @classmethod
   def Get(cls):
@@ -466,7 +621,7 @@ class NewCmd(appcommands.Cmd):
     self._new_style = isinstance(run_with_args, types.MethodType)
     if self._new_style:
       func = run_with_args.im_func
-      code = func.func_code
+      code = func.func_code  # pylint: disable=W0621
       self._full_arg_list = list(code.co_varnames[:code.co_argcount])
       # TODO(user): There might be some corner case where this
       # is *not* the right way to determine bound vs. unbound method.
@@ -631,8 +786,11 @@ class BigqueryCmd(NewCmd):
     retcode = 1
 
     contact_us_msg = (
-        'Please send an email to bigquery-team@google.com to '
-        'report this, with the following information: \n\n')
+        'Google engineers monitor and answer questions on Stack Overflow, with '
+        'the tag google-bigquery:\n'
+        '  http://stackoverflow.com/questions/ask?tags=google-bigquery\n'
+        'Please include a brief description of the steps that led to this '
+        'issue, as well as the following information: \n\n')
     error_details = (
         '========================================\n'
         '== Platform ==\n'
@@ -767,6 +925,11 @@ class _Load(BigqueryCmd):
         'allow_quoted_newlines', None,
         'Whether to allow quoted newlines in CSV import data.',
         flag_values=fv)
+    flags.DEFINE_boolean(
+        'allow_jagged_rows', None,
+        'Whether to allow missing trailing optional columns '
+        'in CSV import data.',
+        flag_values=fv)
     flags.DEFINE_enum(
         'source_format', None,
         ['CSV',
@@ -826,7 +989,7 @@ class _Load(BigqueryCmd):
         'skip_leading_rows': self.skip_leading_rows,
         'max_bad_records': self.max_bad_records,
         'allow_quoted_newlines': self.allow_quoted_newlines,
-        'job_id': FLAGS.job_id,
+        'job_id': _GetJobIdFromFlags(),
         'source_format': self.source_format,
         }
     if self.replace:
@@ -835,7 +998,8 @@ class _Load(BigqueryCmd):
       opts['field_delimiter'] = _NormalizeFieldDelimiter(self.field_delimiter)
     if self.quote is not None:
       opts['quote'] = _NormalizeFieldDelimiter(self.quote)
-
+    if self.allow_jagged_rows is not None:
+      opts['allow_jagged_rows'] = self.allow_jagged_rows
     job = client.Load(table_reference, source, schema=schema, **opts)
     if not FLAGS.sync:
       self.PrintJobStartInfo(job)
@@ -858,6 +1022,30 @@ class _Query(BigqueryCmd):
         'batch', False,
         'Whether to run the query in batch mode.',
         flag_values=fv)
+    flags.DEFINE_boolean(
+        'append_table', False,
+        'When a destination table is specified, whether or not to append.',
+        flag_values=fv)
+    flags.DEFINE_boolean(
+        'replace', False,
+        'If true erase existing contents before loading new data.',
+        flag_values=fv)
+    flags.DEFINE_boolean(
+        'allow_large_results', None,
+        'Enables larger destination table sizes.',
+        flag_values=fv)
+    flags.DEFINE_boolean(
+        'dry_run', None,
+        'Whether the query should be validated without executing.',
+        flag_values=fv)
+    flags.DEFINE_boolean(
+        'require_cache', None,
+        'Whether to only run the query if it is already cached.',
+        flag_values=fv)
+    flags.DEFINE_boolean(
+        'use_cache', None,
+        'Whether to use the query cache to avoid rerunning cached queries.',
+        flag_values=fv)
 
   def RunWithArgs(self, *args):
     """Execute a query.
@@ -870,13 +1058,25 @@ class _Query(BigqueryCmd):
     """
     kwds = {
         'destination_table': self.destination_table,
-        'job_id': FLAGS.job_id,
+        'job_id': _GetJobIdFromFlags(),
+        'allow_large_results': self.allow_large_results,
+        'dry_run': self.dry_run,
+        'use_cache': self.use_cache,
         }
+    if self.destination_table and self.append_table:
+      kwds['write_disposition'] = 'WRITE_APPEND'
+    if self.destination_table and self.replace:
+      kwds['write_disposition'] = 'WRITE_TRUNCATE'
+    if self.require_cache:
+      kwds['create_disposition'] = 'CREATE_NEVER'
     if self.batch:
       kwds['priority'] = 'BATCH'
+
     client = Client.Get()
     job = client.Query(' '.join(args), **kwds)
-    if not FLAGS.sync:
+    if self.dry_run:
+      _PrintDryRunInfo(job)
+    elif not FLAGS.sync:
       self.PrintJobStartInfo(job)
     else:
       _PrintTable(client, job['configuration']['query']['destinationTable'],
@@ -915,7 +1115,7 @@ class _Extract(BigqueryCmd):
     """
     client = Client.Get()
     kwds = {
-        'job_id': FLAGS.job_id,
+        'job_id': _GetJobIdFromFlags(),
         }
     table_reference = client.GetTableReference(source_table)
     job = client.Extract(
@@ -927,10 +1127,14 @@ class _Extract(BigqueryCmd):
 
 
 class _List(BigqueryCmd):
-  usage = """ls [(-j|-p|-d)] [-n <number>] [<identifier>]"""
+  usage = """ls [(-j|-p|-d)] [-a] [-n <number>] [<identifier>]"""
 
   def __init__(self, name, fv):
     super(_List, self).__init__(name, fv)
+    flags.DEFINE_boolean(
+        'all_jobs', None,
+        'Show results from all users in this project (for listing jobs only).',
+        short_name='a', flag_values=fv)
     flags.DEFINE_boolean(
         'jobs', False,
         'Show jobs described by this identifier.',
@@ -968,6 +1172,8 @@ class _List(BigqueryCmd):
           'Cannot specify more than one of -j and -p.')
     if self.p and identifier:
       raise app.UsageError('Cannot specify an identifier with -p')
+    if self.a and not self.j:
+      raise app.UsageError('-a can only be specified with -j')
 
     client = Client.Get()
     formatter = _GetFormatterFromFlags()
@@ -1004,7 +1210,8 @@ class _List(BigqueryCmd):
       results = map(  # pylint:disable-msg=C6402
           client.FormatJobInfo,
           client.ListJobs(reference=project_reference,
-                          max_results=self.max_results))
+                          max_results=self.max_results,
+                          all_users=self.a))
     elif self.p or reference is None:
       BigqueryClient.ConfigureFormatter(formatter, ProjectReference)
       results = map(  # pylint:disable-msg=C6402
@@ -1114,6 +1321,10 @@ class _Copy(BigqueryCmd):
         'force', False,
         "Ignore existing destination tables, don't prompt.",
         short_name='f', flag_values=fv)
+    flags.DEFINE_boolean(
+        'append_table', False,
+        'Append to an existing table.',
+        short_name='a', flag_values=fv)
 
   def RunWithArgs(self, source_table, dest_table):
     """Copies one table to another.
@@ -1125,7 +1336,10 @@ class _Copy(BigqueryCmd):
     source_reference = client.GetTableReference(source_table)
     dest_reference = client.GetTableReference(dest_table)
 
-    if self.no_clobber:
+    if self.append_table:
+      write_disposition = 'WRITE_APPEND'
+      ignore_already_exists = True
+    elif self.no_clobber:
       write_disposition = 'WRITE_EMPTY'
       ignore_already_exists = True
     else:
@@ -1139,7 +1353,7 @@ class _Copy(BigqueryCmd):
     kwds = {
         'write_disposition': write_disposition,
         'ignore_already_exists': ignore_already_exists,
-        'job_id': FLAGS.job_id,
+        'job_id': _GetJobIdFromFlags(),
         }
     job = client.CopyTable(source_reference, dest_reference, **kwds)
     if job is None:
@@ -1352,10 +1566,8 @@ class _Show(BigqueryCmd):
 
     # The JSON formats are handled separately so that they don't print
     # the record as a list of one record.
-    if FLAGS.format == 'prettyjson':
-      print json.dumps(object_info, sort_keys=True, indent=2)
-    elif FLAGS.format == 'json':
-      print json.dumps(object_info, separators=(',', ':'))
+    if FLAGS.format in ['prettyjson', 'json']:
+      _PrintFormattedJsonObject(object_info)
     elif FLAGS.format in [None, 'sparse', 'pretty']:
       formatter = _GetFormatterFromFlags()
       BigqueryClient.ConfigureFormatter(
@@ -1380,7 +1592,7 @@ class _Show(BigqueryCmd):
 
 
 class _Head(BigqueryCmd):
-  usage = """head [-n <max rows>] [<table identifier>]"""
+  usage = """head [-n <max rows>] <table identifier>"""
 
   def __init__(self, name, fv):
     super(_Head, self).__init__(name, fv)
@@ -1398,11 +1610,13 @@ class _Head(BigqueryCmd):
     """
     client = Client.Get()
     reference = client.GetReference(identifier)
-    _Typecheck(reference, (types.NoneType, TableReference),
+    _Typecheck(reference, (TableReference,),
                'Must provide a table identifier for head.')
     _PrintTable(client, dict(reference), max_rows=self.max_rows)
 
 
+
+
 class _Wait(BigqueryCmd):
   usage = """wait [<job_id>] [<secs>]"""
 
@@ -1802,22 +2016,28 @@ class _Version(BigqueryCmd):
 def main(argv):
   try:
     FLAGS.auth_local_webserver = False
+    _ValidateGlobalFlags()
 
-    appcommands.AddCmd('load', _Load)
-    appcommands.AddCmd('query', _Query)
-    appcommands.AddCmd('extract', _Extract)
-    appcommands.AddCmd('ls', _List)
-    appcommands.AddCmd('rm', _Delete)
-    appcommands.AddCmd('mk', _Make)
-    appcommands.AddCmd('show', _Show)
-    appcommands.AddCmd('head', _Head)
-    appcommands.AddCmd('wait', _Wait)
-    appcommands.AddCmd('cp', _Copy)
-    appcommands.AddCmd('update', _Update)
+    bq_commands = {
+        'load': _Load,
+        'query': _Query,
+        'extract': _Extract,
+        'ls': _List,
+        'rm': _Delete,
+        'mk': _Make,
+        'show': _Show,
+        'head': _Head,
+        'wait': _Wait,
+        'cp': _Copy,
+        'update': _Update,
+        'version': _Version,
+        'shell': _Repl,
+        'init': _Init,
+    }
 
-    appcommands.AddCmd('version', _Version)
-    appcommands.AddCmd('shell', _Repl)
-    appcommands.AddCmd('init', _Init)
+    for command, function in bq_commands.iteritems():
+      if command not in appcommands.GetCommandList():
+        appcommands.AddCmd(command, function)
 
     if (not argv or
         (len(argv) > 1 and
@@ -1825,7 +2045,7 @@ def main(argv):
          argv[1] in appcommands.GetCommandList())):
       # Service Accounts don't use cached oauth credentials and
       # all bigqueryrc defaults are technically optional.
-      if not FLAGS.service_account:
+      if not _UseServiceAccount():
         if not (os.path.exists(_GetBigqueryRcFilename()) or
                 os.path.exists(FLAGS.credential_file)):
           appcommands.GetCommandByName('init').Run([])
